/*-
 * Copyright (c) 2000 Russell Vincent
 * All rights reserved.
 * Algorithms Copyright (c) 2000 Joe Greco and sol.net Network Services
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 *
 *  Try and categorise an article into a range of types
 *
 *  We keep state info as the article is passed to us line by line
 *
 *  Once we have found an article type, we keep to that type unless
 *  we find a different type. We never reset.
 *
 *  Once we have found a binary, we stop scanning the article - save CPU
 *
 */

#include "defs.h"

Prototype void InitArticleType(void);
Prototype int ArtTypeConv(char *t);
Prototype int ArtTypeMatch(int articletype, ArtTypeList *acctypes);
Prototype int ArticleType(char *line, int len, int inheader);

int isBinHexLine(char *line, int len, int firstline);
int isUuencodedLine(char *buf, int len);

/*
 * Global state variables required for multi-line checking
 */
int Type = ARTTYPE_DEFAULT;
int UUencode = 0;
int BinHex = 0;
int Base64 = 0;

char validBinHex[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 0-15 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 16-31 */
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0,		/* 32-47 */
	1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0,		/* 48-63 */
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0,		/* 64-79 */
	1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0,		/* 80-95 */
	1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0,		/* 96-111 */
	1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 112-127 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 128- */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

char validBase64[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 0-15 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,		/* 16-31 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1,		/* 32-47 */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0,		/* 48-63 */
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		/* 64-79 */
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,		/* 80-95 */
	0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		/* 96-111 */
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		/* 112-127 */
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,		/* 128- */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
};

void
InitArticleType(void)
{
    Type = ARTTYPE_DEFAULT;
    UUencode = 0;
    BinHex = 0;
    Base64 = 0;
}

int
ArtTypeConv(char *t)
{
    if (strcasecmp(t, "none") == 0)
	return(ARTTYPE_NONE);
    else if (strcasecmp(t, "default") == 0)
	return(ARTTYPE_DEFAULT);
    else if (strcasecmp(t, "control") == 0)
	return(ARTTYPE_CONTROL);
    else if (strcasecmp(t, "cancel") == 0)
	return(ARTTYPE_CANCEL);
    else if (strcasecmp(t, "mime") == 0)
	return(ARTTYPE_MIME);
    else if (strcasecmp(t, "binary") == 0 || strcasecmp(t, "binaries") == 0)
	return(ARTTYPE_BINARY);
    else if (strcasecmp(t, "uuencode") == 0)
	return(ARTTYPE_UUENCODE);
    else if (strcasecmp(t, "base64") == 0)
	return(ARTTYPE_BASE64);
    else if (strcasecmp(t, "multipart") == 0)
	return(ARTTYPE_MULTIPART);
    else if (strcasecmp(t, "html") == 0)
	return(ARTTYPE_HTML);
    else if (strcasecmp(t, "ps") == 0)
	return(ARTTYPE_PS);
    else if (strcasecmp(t, "binhex") == 0)
	return(ARTTYPE_BINHEX);
    else if (strcasecmp(t, "partial") == 0)
	return(ARTTYPE_PARTIAL);
    else if (strcasecmp(t, "pgp") == 0)
	return(ARTTYPE_PGPMESSAGE);
    else if (strcasecmp(t, "all") == 0)
	return(ARTTYPE_ALL);
    else {
	syslog(LOG_ERR, "Unknown article type: %s", t);
	return(ARTTYPE_NONE);
    }
}

/*
 * Match a article type to a list of types that are acceptable
 *
 * Returns:  1 = Match
 *	     0 = No match
 *
 */
int
ArtTypeMatch(int articletype, ArtTypeList *acctypes)
{
    int res = 0;
    if (acctypes == NULL)
	return(1);
    if (acctypes->negate)
	res = 1;
    while (acctypes != NULL) {
	if ((articletype & acctypes->arttype) == acctypes->arttype)
		res = !acctypes->negate;
	acctypes = acctypes->next;
    }
    return(res);
}

int
isBinHexLine(char *line, int len, int firstline)
{
    int count;
    char *p = line;

    if (len != 64)
	return(0);
    for (count = 0; count < 64; count++) {
	if (!validBinHex[(int)*p]) {
	    if (firstline && (char *)p == line && *p == ':')
		;		/* no action */
	    else
		return(0);
	}
	p++;
    }
    if (*p != '\n')
	return(0);
    return(1);
}

int
isUuencodedLine(char *buf, int len)
{
    char *ptr, *end;
    int count;

    end = buf + len;
    ptr = buf;

    if (len != 61)
	return(0);
    if (*ptr != 'M')
	return(0);
    ptr++;
    for (count = 0; count < 60; count++) {
	if (*ptr < ' ' || *ptr >= 'a')
	    return(0);
	ptr++;
    }
    if (*ptr != '\n')
	return(0);
    return(1);
}

int
isBase64(char *p, int len)
{
    int count;

    if (len != 74)
	return(0);

    for (count = 0; count < 74; count++) {
	if (!validBase64[(int)*p])
	    return(0);
	p++;
    }
    return(1);
}

int
ArticleType(char *line, int len, int inheader)
{
    static char *ptr;
    static char first;

    if (Type != ARTTYPE_DEFAULT)
	Type = Type & ~ARTTYPE_DEFAULT;
    /*
     * If we have found a binary, just keep it that way - saves CPU
     */
    if (Type & ARTTYPE_BINARY)
	return(Type);

    ptr = line;

    /*
     * Do some checks that could apply to headers and or body
     */
    first = tolower(*ptr);
    if (first == 'c' && tolower(*(ptr + 1) == 'o')) {
	if (! strncasecmp(ptr, "Content-Type: text/html", 23))
	    Type |= ARTTYPE_HTML;
	if (! strncasecmp(ptr, "Content-Type: ", 14))
	    Type |= ARTTYPE_MIME;
	if (! strncasecmp(ptr, "Content-Type: multipart", 23))
	    Type |= ARTTYPE_MULTIPART;
	if (! strncasecmp(ptr, "Content-transfer-encoding: base64", 33))
	    Type |= ARTTYPE_BASE64 | ARTTYPE_BINARY;
	if (! strncasecmp(ptr, "Content-Type: application/postscript", 36))
	    Type |= ARTTYPE_PS;
	if (! strncasecmp(ptr, "Content-Type: application/mac-binhex40", 38))
	    BinHex = 1;
	if (! strncasecmp(ptr, "Content-Type: message/partial", 29))
	    Type |= ARTTYPE_PARTIAL;
    }
    if (first == '(' && strncasecmp(ptr, "(This file must be converted with BinHex 4.0)", 33) == 0)
	BinHex = 1;

    if (inheader) {
	if (first == 'c' && tolower(*(ptr + 1) == 'o')) {
	    if (! strncasecmp(ptr, "Control: ", 9))
		Type |= ARTTYPE_CONTROL;
	    if (! strncasecmp(ptr, "Control: cancel ", 16))
		Type |= ARTTYPE_CANCEL;
	}
	if (first == 'm' && strncasecmp(ptr, "Mime-Version: ", 14) == 0)
		Type |= ARTTYPE_MIME;
    } else {
	/* Get quoted UUencode, etc. */
	if (*ptr == '>') {
	    ptr++;
	    if (*ptr == ' ')
		ptr++;
	}

	/*
	 * We look for binary formats first, since these eat 
	 * up CPU like there is no tomorrow.  We may miss 
	 * some flags if we detect a binary file, the 
	 * alternative is to remove the return(Type)'s and 
	 * eat CPU for the entire binary.
	 */

	/* Check UUencode */
	if (len == 61 && *ptr == 'M') {
	    if (isUuencodedLine(ptr, len)) {
		UUencode++;
		if (UUencode > 8) {
		    Type |= ARTTYPE_UUENCODE | ARTTYPE_BINARY;
		    return(Type);
		}
	    } else {
		UUencode = 0;
	    }
	} else if (len == 64) {
	    /* Check BinHex */
	    if (isBinHexLine(ptr, len, BinHex < 2 ?  1 : 0)) {
		BinHex++;
		if (BinHex > 8) {
		    Type |= ARTTYPE_BINHEX | ARTTYPE_BINARY;
		    return(Type);
		}
	    } else {
		BinHex = 0;
	    }
	} else if (len == 74) {
	    if (isBase64(ptr, len)) {
		Base64++;
		if (Base64 > 8) {
		    Type |= ARTTYPE_BASE64 | ARTTYPE_BINARY;
		    return(Type);
		}
	    } else {
		Base64 = 0;
	    }
	} else if (*ptr == '-' &&
			strncmp(ptr, "-----BEGIN PGP MESSAGE-----", 27) == 0)
	    Type |= ARTTYPE_PGPMESSAGE;
    }
    if (Type != ARTTYPE_DEFAULT)
	Type = Type & ~ARTTYPE_DEFAULT;
    return(Type);
}


